/***************************************************************************
 *   Copyright (C) 2005 by Roberto Cappuccio                               *
 *   robertoc@dnet.it                                                      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "katdatabasemanager.h"

#include <iostream>
#include <qregexp.h>
#include <qbuffer.h>
#include <qfile.h>
#include <kdebug.h>

using namespace std;


KatDatabaseManager::KatDatabaseManager(DlgMainWindow *_dm) {
    db = new CppSQLite3DB();
    dm = _dm;
}

KatDatabaseManager::~KatDatabaseManager() {
    kdDebug()<<"KatDatabaseManager::~KatDatabaseManager() \n";
    delete db;
}

int KatDatabaseManager::createKatDatabase( const QString &filename ) {

    try {
        QString dbfilename = filename;
        QString journalfilename = filename + "-journal";

        QFile::remove( dbfilename );
        QFile::remove( journalfilename );

        db->open( dbfilename );

        // create the tables
        db->execDML( "create table catalog("
                                            "name varchar(100) primary key, "
                                            "description varchar(500), "
                                            "filename varchar(500), "
                                            "path varchar(1000), "
                                            "notes varchar(2000), "
                                            "author varchar(100), "
                                            "thumbnailsize integer, "
                                            "useexclusionlist integer, "
                                            "creationdate varchar(12), "
                                            "lastupdatedate varchar(12), "
                                            "metadata integer, "
                                            "files integer, "
                                            "folders integer, "
                                            "fulltexts integer, "
                                            "thumbnails integer, "
                                            "words integer"
                                            ");" );


        db->execDML( "create table files("
                                            "fileid integer primary key, "
                                            "fullname varchar(5000), "
                                            "filename varchar(300), "
                                            "parentid integer, "
                                            "filetype varchar(100), "
                                            "filesize integer, "
                                            "statuschangedate integer, "
                                            "modificationdate integer, "
                                            "lastaccessdate integer, "
                                            "username varchar(100),"
                                            "groupname varchar(100),"
                                            "permissions integer,"
                                            "mode integer,"
                                            "language varchar(100)"
                                            ");" );

        db->execDML( "create index files_filename on files (filename);" );
        db->execDML( "create unique index files_fullname on files (fullname);" );

        db->execDML( "create table words("
                                            "word varchar(100), "
                                            "wordid integer primary key"
                                            ");" );

        db->execDML( "create unique index words_word on words (word);" );
        db->execDML( "create unique index words_wordid on words (wordid);" );

        db->execDML( "create table wordfile("
                                            "wordid integer, "
                                            "fileid integer, "
                                            "occurrences integer "
                                            ");" );

        db->execDML( "create index wordfile_wordid on wordfile (wordid);" );
        db->execDML( "create index wordfile_fileid on wordfile (fileid);" );


        db->execDML( "create table parents("
                                            "parentid integer, "
                                            "childid integer"
                                            ");" );

        db->execDML( "create index parents_parentid on parents (parentid);" );
        db->execDML( "create index parents_childid on parents (childid);" );


        db->execDML( "create table metadata("
                                            "fileid integer, "
                                            "field varchar(100), "
                                            "type varchar(100), "
                                            "value varchar(5000)"
                                            ");" );

        db->execDML( "create index metadata_fileid on metadata (fileid);" );
        db->execDML( "create index metadata_field on metadata (field);" );
        db->execDML( "create index metadata_type on metadata (type);" );
        db->execDML( "create index metadata_value on metadata (value);" );


        db->execDML( "create table thumbnails("
                                            "fileid integer primary key, "
                                            "thumbnaildata blob, "
                                            "thumbnaildatalength integer "
                                            ");" );

        db->execDML( "create table fulltexts("
                                            "fileid integer primary key, "
                                            "fulltextdata blob, "
                                            "fulltextdatalength integer "
                                            ");" );

    } catch ( CppSQLite3Exception& e ) {

        cerr << e.errorCode() << ":" << e.errorMessage() << endl;

        return e.errorCode();
    }

    return 0;
}

int KatDatabaseManager::saveCatalogInformation( KatCatalog& kat ) {

    CppSQLite3Statement stmt = db->compileStatement( "insert into catalog "
                                                     "values(?,?,?,?,?,?,?,?,?,?,?,?,?,?,?,?);" );
    stmt.bind( 1, kat.getName() );
    stmt.bind( 2, kat.getDescription() );
    stmt.bind( 3, kat.getFilename() );
    stmt.bind( 4, kat.getPath() );
    stmt.bind( 5, kat.getNotes() );
    stmt.bind( 6, kat.getAuthor() );
    stmt.bind( 7, kat.getThumbnailSize() );
    stmt.bind( 8, kat.getUseExclusionList() );
    stmt.bind( 9, kat.getCreationDate() );
    stmt.bind(10, kat.getLastUpdateDate() );
    stmt.bind(11, kat.getMetaData() );
    stmt.bind(12, kat.getFiles() );
    stmt.bind(13, kat.getFolders() );
    stmt.bind(14, kat.getFullTexts() );
    stmt.bind(15, kat.getThumbnails() );
    stmt.bind(16, kat.getWords() );

    try {

        stmt.execDML();
        stmt.finalize();

    } catch ( CppSQLite3Exception& e ) {

        cerr << e.errorCode() << ":" << e.errorMessage() << endl;

        return e.errorCode();
    }

    return 0;
}

int KatDatabaseManager::readCatalogInformation( KatCatalog& kat ) {

    QString DML;

    // insert the word into the words table, if not present
    DML = "select * from catalog;";

    try {
        CppSQLite3Query q = db->execQuery( DML );
        if ( !q.eof() ) {
            kat.setName( q.getStringField( "name" ) );
            kat.setDescription( q.getStringField( "description" ) );
            kat.setFilename( q.getStringField( "filename" ) );
            kat.setPath( q.getStringField( "path" ) );
            kat.setNotes( q.getStringField( "notes" ) );
            kat.setAuthor( q.getStringField( "author" ) );
            kat.setThumbnailSize( q.getIntField( "thumbnailsize" ) );
            kat.setUseExclusionList( q.getIntField( "useexclusionlist" ) );
            kat.setCreationDate( q.getIntField( "creationdate" ) );
            kat.setLastUpdateDate( q.getIntField( "lastupdatedate" ) );
            kat.setMetaData( q.getIntField( "metadata" ) );
            kat.setFiles( q.getIntField( "files" ) );
            kat.setFolders( q.getIntField( "folders" ) );
            kat.setFullTexts( q.getIntField( "fulltexts" ) );
            kat.setThumbnails( q.getIntField( "thumbnails" ) );
            kat.setWords( q.getIntField( "words" ) );
        }
        q.finalize();

    } catch ( CppSQLite3Exception& e ) {

        cerr << e.errorCode() << ":" << e.errorMessage() << endl;

        return e.errorCode();
    }

    return 0;
}

long KatDatabaseManager::saveFileRecord( KFileItem* kfi ) {

    QString DML;
    long parentId;

    QString fullName = kfi->url().path() ;
    unsigned pos = fullName.findRev( '/', -1 );
    QString fileName = fullName.right( fullName.length() - pos - 1 );
    QString path = fullName.left( pos );

    // TODO: Find a regular expression to eliminate all possible problems with filenames in SQL
    QString spath( path );
    spath = spath.replace( QRegExp("'"), "''" );

    // Find the parent of current file
    try {
        DML = "select fileid from files where fullname = '" + spath + "';";
        CppSQLite3Query q = db->execQuery( DML );

        if ( !q.eof() ) {
            parentId = q.getIntField( "fileid" );
        } else {
            parentId = -1L;
        }
        q.finalize();

    } catch ( CppSQLite3Exception& e ) {

        cerr << e.errorCode() << ":" << e.errorMessage() << endl;

        return 0;
    }

    // Save the basic information into the files table
    try {
        CppSQLite3Statement stmt = db->compileStatement( "insert into files("
                                                            "fullname,"
                                                            "filename,"
                                                            "parentid,"
                                                            "filetype,"
                                                            "filesize,"
                                                            "statuschangedate,"
                                                            "modificationdate,"
                                                            "lastaccessdate,"
                                                            "username,"
                                                            "groupname,"
                                                            "permissions,"
                                                            "mode,"
                                                            "language"
                                                          ") values (?,?,?,?,?,?,?,?,?,?,?,?,?);"
                                                        );

        stmt.bind(  1, kfi->url().path() );
        stmt.bind(  2, fileName );
        stmt.bind(  3, parentId );
        stmt.bind(  4, kfi->mimetype() );
        stmt.bind(  5, (long)kfi->size() );
        stmt.bind(  6, kfi->time( KIO::UDS_CREATION_TIME ) );
        stmt.bind(  7, kfi->time( KIO::UDS_MODIFICATION_TIME ) );
        stmt.bind(  8, kfi->time( KIO::UDS_ACCESS_TIME ) );
        stmt.bind(  9, kfi->user() );
        stmt.bind( 10, kfi->group() );
        stmt.bind( 11, kfi->permissions() );
        stmt.bind( 12, kfi->mode() );
        stmt.bind( 13, QString::null );

        stmt.execDML();
        stmt.finalize();

    } catch ( CppSQLite3Exception& e ) {

        cerr << e.errorCode() << ":" << e.errorMessage() << " : " << kfi->url().path() << endl;

        return 0;
    }

    long fileId = db->lastRowId();

    // Save parent information into the parents table
    try {

        CppSQLite3Statement stmt = db->compileStatement( "insert into parents values(?,?);" );
        stmt.bind( 1, parentId );
        stmt.bind( 2, fileId );

        stmt.execDML();
        stmt.finalize();

    } catch ( CppSQLite3Exception& e ) {

        cerr << e.errorCode() << ":" << e.errorMessage() << endl;

        return e.errorCode();
    }

    return fileId;
}

int KatDatabaseManager::saveThumbnailRecord( int fileId, QByteArray& thumbnailData, long thumbnailDataLength ) {

    // Save the thumbnail into the thumbnails table
    try {
        CppSQLite3Statement stmt = db->compileStatement( "insert into thumbnails values(?,?,?);" );
        stmt.bind(  1, fileId );
        stmt.bind(  2, (unsigned char*)thumbnailData.data(), thumbnailDataLength );
        stmt.bind(  3, thumbnailDataLength );

        stmt.execDML();
        stmt.finalize();

    } catch ( CppSQLite3Exception& e ) {

        cerr << e.errorCode() << ":" << e.errorMessage() << endl;

        return e.errorCode();
    }

    return 0;
}

int KatDatabaseManager::saveMetaDataRecord( int fileId, const QString &metaData ) {

    // Tokenize the metaData string and insert it into the metadata table
    if ( !metaData.isNull() ) {
        QString name;
        QString type;
        QString value;

        QStringList meta = QStringList::split( "|", metaData, false );
        for ( QStringList::Iterator it = meta.begin(); it != meta.end(); it++ ) {
            name = *it;
            it++;
            type = *it;
            it++;
            value = *it;

            try {
                CppSQLite3Statement stmt = db->compileStatement( "insert into metadata values(?,?,?,?);" );
                stmt.bind(  1, fileId );
                stmt.bind(  2, name );
                stmt.bind(  3, type );
                stmt.bind(  4, value );

                stmt.execDML();
                stmt.finalize();

            } catch ( CppSQLite3Exception& e ) {

                cerr << e.errorCode() << ":" << e.errorMessage() << endl;

                return e.errorCode();
            }
        }
    }

    return 0;
}

int KatDatabaseManager::saveFullTextRecord( int fileId, QByteArray& fullTextData, long fullTextDataLength ) {

    // Save the fulltext into the fulltexts table
    try {
        CppSQLite3Statement stmt = db->compileStatement( "insert into fulltexts values(?,?,?);" );
        stmt.bind(  1, fileId );
        stmt.bind(  2, (unsigned char*)fullTextData.data(), fullTextDataLength );
        stmt.bind(  3, fullTextDataLength );

        stmt.execDML();
        stmt.finalize();

    } catch ( CppSQLite3Exception& e ) {

        cerr << e.errorCode() << ":" << e.errorMessage() << endl;

        return e.errorCode();
    }

    return 0;
}

int KatDatabaseManager::updateLanguageOfFile( int fileId, QString& language ) {

    // Update the language information in the file record
    try {
        db->execDML( "update files set language = '" + language + "' where fileid=" + QString::number( fileId ) + ";" );

    } catch ( CppSQLite3Exception& e ) {

        cerr << e.errorCode() << ":" << e.errorMessage() << endl;

        return e.errorCode();
    }

    return 0;
}

QMap<long,QString> KatDatabaseManager::getFileChildren( long parentId ) {

    QMap<long, QString> children;

    QString DML;

    try {
        DML = "select childid, fullname from (parents inner join files on parents.childid = files.fileid) "
              "where parentid = '" + QString::number( parentId ) + "';";

        CppSQLite3Query q = db->execQuery( DML );

        while ( !q.eof() ) {
            long fileId = q.getIntField( "childid" );
            QString fullName = q.getStringField( "fullname" );

            children.insert( fileId, fullName );

            q.nextRow();
        }
        q.finalize();

    } catch ( CppSQLite3Exception& e ) {


        return children;
    }

    return children;
}

KatInformation KatDatabaseManager::readFileInformation( long fileId ) {

    KatInformation info;

    QString DML;

    // Read the basic information present in the files table
    try {
        DML = "select * from files where fileid = '" + QString::number( fileId ) + "';";

        CppSQLite3Query q = db->execQuery( DML );

        if ( !q.eof() ) {

            info.fileId = q.getIntField( "fileid" );
            info.fullName = q.getStringField( "fullname" );
            info.fileName = q.getStringField( "filename" );
            info.parentId = q.getIntField( "parentid" );
            info.fileType = q.getStringField( "filetype" );
            info.fileSize = q.getIntField( "filesize" );
            info.statusChangeDate = q.getIntField( "statuschangedate" );
            info.modificationDate = q.getIntField( "modificationdate" );
            info.lastAccessDate = q.getIntField( "lastaccessdate" );
            info.language = q.getStringField( "language" );

            // TODO: Add the new fields (group, user, permissions...)
        }
        q.finalize();

    } catch ( CppSQLite3Exception& e ) {

        cerr << e.errorCode() << ":" << e.errorMessage() << endl;

        return info;
    }


    // Read the thumbnail present in the thumbnails table
    try {
        DML = "select * from thumbnails where fileid = '" + QString::number( fileId ) + "';";

        CppSQLite3Query q = db->execQuery( DML );

        if ( !q.eof() ) {

            info.thumbnailDataLength = q.getIntField( "thumbnaildatalength" );
            info.hasThumbnail = true;

            QBuffer buffer( info.thumbnailData );
            buffer.open( IO_WriteOnly );
            int len = (int)info.thumbnailDataLength;
            buffer.writeBlock( (char*)q.getBlobField( "thumbnaildata" , len ), (long)len );
            buffer.close();
        }
        q.finalize();

    } catch ( CppSQLite3Exception& e ) {

        cerr << e.errorCode() << ":" << e.errorMessage() << endl;

        return info;
    }

    // Read the metadata information present in the metadata table
    QString metaData = "";
    try {
        DML = "select * from metadata where fileid = '" + QString::number( fileId ) + "';";

        CppSQLite3Query q = db->execQuery( DML );

        while ( !q.eof() ) {

            info.hasMetaData = true;

            QString field = q.getStringField( "field" );
            QString type = q.getStringField( "type" );
            QString value = q.getStringField( "value" );

            metaData = metaData + field + "|" + type + "|" + value + "|";

            q.nextRow();
        }
        q.finalize();

        info.metaData = metaData;

    } catch ( CppSQLite3Exception& e ) {

        cerr << e.errorCode() << ":" << e.errorMessage() << endl;

        return info;
    }

    // Check if the file has a fulltext record
    try {
        DML = "select fileid from fulltexts where fileid = '" + QString::number( fileId ) + "';";

        CppSQLite3Query q = db->execQuery( DML );

        if ( !q.eof() ) {
            info.hasFullText = true;
        }
        q.finalize();

    } catch ( CppSQLite3Exception& e ) {

        cerr << e.errorCode() << ":" << e.errorMessage() << endl;

        return info;
    }

    return info;
}

blobInfo KatDatabaseManager::readFullTextRecord( long fileId ) {

    QString DML;
    blobInfo bi;
    bi.blobLen = 0;

    // Read the fulltext present in the fulltext table
    try {
        DML = "select * from fulltexts where fileid = '" + QString::number( fileId ) + "';";

        CppSQLite3Query q = db->execQuery( DML );

        if ( !q.eof() ) {

            bi.blobLen = q.getIntField( "fulltextdatalength" );

            QBuffer buffer( bi.blobData );
            buffer.open( IO_WriteOnly );
            int len = bi.blobLen;
            buffer.writeBlock( (char*)q.getBlobField( "fulltextdata" , len ), (long)len );
            buffer.close();
        }
        q.finalize();

    } catch ( CppSQLite3Exception& e ) {

        cerr << e.errorCode() << ":" << e.errorMessage() << endl;

        return bi;
    }

    return bi;
}

QValueList<int> KatDatabaseManager::doSearch( const QString &what, bool filenames, bool descriptions, bool fulltexts ) {

    QValueList<int> results;
    QString DML = "";

    // TODO: Find a regular expression to eliminate all possible problems with words in SQL
    QString swhat( what );
    swhat = swhat.replace( QRegExp("'"), "''" );

    try {

        if ( fulltexts ) {
            DML = DML + "select distinct fileid from wordfile join words on wordfile.wordid = words.wordid "
                        "where word like '%" + swhat + "%'";

            if ( filenames || descriptions ) {
                DML = DML + " union ";
            }
/*
            DML = DML + "select distinct fileid from fulltexts "
                        "where fulltextdata like '%" + swhat + "%'";

            if ( filenames || descriptions ) {
                DML = DML + " union ";
            }
*/
        }

        if ( filenames ) {
            DML = DML + "select fileid from files where filename like '%" + swhat + "%'";

            if ( descriptions ) {
                DML = DML + " union ";
            }
        }

        if ( descriptions ) {
            DML = DML + "select fileid from metadata where value like '%" + swhat + "%'";
        }

        DML = DML + ";";

        //cout << DML << endl;

        CppSQLite3Query q = db->execQuery( DML );

        while ( !q.eof() ) {
            int fileId = q.getIntField( "fileid" );
            results.push_back( fileId );

            q.nextRow();
        }
        q.finalize();

    } catch ( CppSQLite3Exception& e ) {

        cerr << e.errorCode() << ":" << e.errorMessage() << endl;

        return results;
    }

    return results;
}

int KatDatabaseManager::openTransaction() {

    try {

        db->execDML( "begin transaction;" );

    } catch ( CppSQLite3Exception& e ) {

        return e.errorCode();
    }

    return 0;
}

int KatDatabaseManager::commitTransaction() {

    try {

        db->execDML( "commit transaction;" );

    } catch ( CppSQLite3Exception& e ) {

        return e.errorCode();
    }

    return 0;
}

int KatDatabaseManager::openKatDatabase( const QString& filename ) {

    struct stat filestat;

    if ( stat( filename, &filestat ) == -1 ) {

        return -1;

    }

    try {

        db->open( filename );

    } catch ( CppSQLite3Exception& e ) {

        return e.errorCode();
    }

    return 0;
}

int KatDatabaseManager::closeKatDatabase() {

    try {

        db->close();

    } catch ( CppSQLite3Exception& e ) {

        return e.errorCode();
    }

    return 0;
}

long KatDatabaseManager::getLastWordId() {
    QString DML;
    long lastWordId;

    DML = "select max(wordid) as lastwordid from words;";
    CppSQLite3Query q = db->execQuery( DML );
    if ( q.eof() ) {
        // the words table is empty
        lastWordId = 0;
    } else {
        lastWordId = q.getIntField( "lastwordid" );
    }

    q.finalize();

    return lastWordId;
}

QMap<QString, long> KatDatabaseManager::readFileWords( long fileId ) {
    QString DML;
    QMap<QString, long> words;



    if ( fileId == -1L ) {
        // read all words from the catalog

    } else {
        // read the words found in the specified file
        try {
            openTransaction();

            DML = "select wordfile.occurrences, words.word from wordfile "
                        "join words on wordfile.wordid = words.wordid "
                        "where wordfile.fileid=" + QString::number( fileId ) + ";";

            CppSQLite3Query q = db->execQuery( DML );
            while ( !q.eof() ) {

                words[ q.getStringField( "word" ) ] = q.getIntField( "occurrences" );

                q.nextRow();
            }
            q.finalize();

            commitTransaction();

        } catch ( CppSQLite3Exception& e ) {

            cerr << e.errorCode() << ":" << e.errorMessage() << endl;

            return words;
        }
    }

    return words;
}

long KatDatabaseManager::saveWordRecord( QString &word ) {
    QString DML;

    // TODO: Find a regular expression to eliminate all possible problems with words in SQL
    QString sword( word );
    sword = sword.replace( QRegExp("'"), "''" );

    try {
        // insert the word into the words table
        CppSQLite3Statement stmt = db->compileStatement( "insert into words (word) values(?);" );
        stmt.bind( 1, word );

        stmt.execDML();
        stmt.finalize();

    } catch ( CppSQLite3Exception& e ) {

        cerr << e.errorCode() << ":" << e.errorMessage() << " WORD WAS: " << sword << " WORDID: " << db->lastRowId() << endl;

        return -1;

    }

    return db->lastRowId();
}

int KatDatabaseManager::saveWordfileRecord( long wordId, long fileId, long wordCount ) {
    QString DML;

    try {
        // insert the wordfile information into the wordfile table
        CppSQLite3Statement stmt = db->compileStatement( "insert into wordfile values(?,?,?);" );
        stmt.bind( 1, wordId );
        stmt.bind( 2, fileId );
        stmt.bind( 3, wordCount );

        stmt.execDML();
        stmt.finalize();

    } catch ( CppSQLite3Exception& e ) {

        cerr << e.errorCode() << ":" << e.errorMessage() << endl;

        return e.errorCode();

    }

    return 0;
}

long KatDatabaseManager::getWordId( QString& word ) {
    QString DML;
    long wordId;

    // TODO: Find a regular expression to eliminate all possible problems with words in SQL
    QString sword( word );
    sword = sword.replace( QRegExp("'"), "''" );

    DML = "select wordid from words where word = '" + sword + "';";
    CppSQLite3Query q = db->execQuery( DML );
    if ( q.eof() ) {

        // word absent
        wordId = -1;

    } else {

        // word already present
        wordId = q.getIntField( "wordid" );

    }
    q.finalize();

    return wordId;
}

bool KatDatabaseManager::folderHasChildren( long fileId ) {

    long childrenNumber = 0L;

    try {
        childrenNumber = db->execScalar( "select count(*) from parents join files on parents.childid = files.fileid "
                                              "where parents.parentid = " + QString::number( fileId ) + " and files.filetype = 'inode/directory';");

    } catch ( CppSQLite3Exception& e ) {

        cerr << e.errorCode() << ":" << e.errorMessage() << endl;

        return false;
    }

    if ( childrenNumber != 0 ) {
        return true;
    } else {
        return false;
    }
}

QMap<QString, long> KatDatabaseManager::readFiles( long offset, long limit ) {
    QString DML;
    QMap<QString, long> files;

    // read the words found in the specified file
    try {
        DML = "select fileid, fullname from files limit " + QString::number( limit ) +
              " offset " + QString::number( offset ) + ";";

        CppSQLite3Query q = db->execQuery( DML );
        while ( !q.eof() ) {
            files[ q.getStringField( "fullname" ) ] = q.getIntField( "fileid" );
            q.nextRow();
        }
        q.finalize();

    } catch ( CppSQLite3Exception& e ) {

        cerr << e.errorCode() << ":" << e.errorMessage() << endl;

        return files;
    }

    return files;
}
